@@ -2,7 +2,7 @@
   This file is part of shuJIT,
   Just In Time compiler for Sun Java Virtual Machine.
 
-  Copyright (C) 1998,1999 SHUDO Kazuyuki
+  Copyright (C) 1998,1999,2000 SHUDO Kazuyuki
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -21,50 +21,41 @@
   $Id$
 */
 
-#include <stdlib.h>	/* for getenv(),atoi() */
-#include <sys/stat.h>	/* for gdbm_open(), stat() */
-#include <unistd.h>	/* for stat() */
+#include <stdlib.h>	// for getenv(),atoi()
+#include <sys/stat.h>	// for gdbm_open(), stat()
+#include <unistd.h>	// for stat()
+#include <string.h>	// for memset()
 
 #include "compiler.h"
 
-#if JDK_VER >= 12
-#  include "jit.h"	/* for JITInterface, JITInterface6 */
-#endif
-
 #if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
 #  include <signal.h>
-#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */
+#endif	// EXC_BY_SIGNAL || GET_SIGCONTEXT
+
+#if JDK_VER >= 12
+#  include "jit.h"	// for JITInterface, JITInterface6
+#endif
 
 #ifdef CODE_DB
-#  include <fcntl.h>	/* for O_... */
-#  include <dlfcn.h>	/* for dl*() */
-#endif	/* CODE_DB */
+#  include <fcntl.h>	// for O_...
+#  include <dlfcn.h>	// for dl*()
+#endif	// CODE_DB
 
 #ifdef METAVM
 #  include "jni.h"
-#  include "metavm/metavm.h"	/* for macro METAVM_PKG */
+#  include "metavm/metavm.h"	// for macro METAVM_PKG
 #endif
 
 
-/*
- * Global Variables
- */
+//
+// Global Variables
+//
 #ifndef IGNORE_DISABLE
 bool_t compiler_enabled = TRUE;
-#endif	/* IGNORE_DISABLE */
+#endif	// IGNORE_DISABLE
 
-#if (defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)) && defined(SEARCH_SIGCONTEXT)
-#  ifdef linux
-unsigned short reg_gs;
-unsigned short reg_fs;
-#  endif
-unsigned short reg_es;
-unsigned short reg_ds;
-#  ifdef __FreeBSD__
-unsigned short reg_cs;
-unsigned short reg_ss;
-#  endif
-#endif	/* (EXC_BY_SIGNAL || GET_SIGCONTEXT) && SEARCH_SIGCONTEXT */
+// for strictfp
+bool_t is_fpupc_double = FALSE;
 
 #ifdef CODE_DB
 #  ifdef GDBM
@@ -75,7 +66,7 @@ GDBM_FILE db = NULL;
 DBM *db = NULL;
 #  endif
 int db_page = -1;
-#endif	/* CODE_DB */
+#endif	// CODE_DB
 
 struct bool_opt_entry { char *name; int num; };
 struct bool_opt_entry bool_opt_entry[] = {
@@ -86,8 +77,11 @@ struct bool_opt_entry bool_opt_entry[] = {
   {"igndisable", OPT_IGNDISABLE},
 #ifdef CODE_DB
   {"codedb", OPT_CODEDB},
-#endif	/* CODE_DB */
+#endif	// CODE_DB
   {"cmplatload", OPT_CMPLATLOAD},
+  {"ignlock", OPT_IGNLOCK},
+  {"ignstrictfp", OPT_IGNSTRICTFP},
+  {"frcstrictfp", OPT_FRCSTRICTFP},
   {NULL, -1}
 };
 int options = 0;
@@ -112,40 +106,30 @@ void (*sym_dbm_close)(DBM *);
 int (*sym_dbm_store)(DBM *,datum,datum,int);
 datum (*sym_dbm_fetch)(DBM *,datum);
 #  endif
-#endif	/* CODE_DB */
-
-#ifdef EXC_BY_SIGNAL
-static struct sigaction sigActForHandler;
-#endif
+#endif	// CODE_DB
 
 
-/*
- * Local Functions
- */
-static void initializeClass(ClassClass *);
-#ifdef __FreeBSD__
-static bool_t signalHandler(int sig, int code, struct sigcontext *uc);
-#else
-static bool_t signalHandler(int sig, void *info, void *uc);
-#endif
-#define fillFPUEnv(FPUENV)	asm("fstenv %0" : "=m" (FPUENV))
-static void showFPUEnv(FILE *, char *indent, char fpuenv[28]);
+//
+// Local Functions
+//
+static void initializeClassHook(ClassClass *);
 static void freeClass(ClassClass *);
 static unsigned char *compiledCodePC(JavaFrame *, struct methodblock *);
 static bool_t pcInCompiledCode(unsigned char *pc, struct methodblock *);
+#ifdef DIRECT_INVOCATION
 static JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf);
+#endif	// DIRECT_INVOCATION
+#if JDK_VER >= 12
+static void *frameID(JavaFrame *frame);
+#endif
+#ifndef IGNORE_DISABLE
 static void compilerEnable();
 static void compilerDisable();
+#endif	// IGNORE_DISABLE
 static bool_t compileClass(ClassClass *);
 static bool_t compileClasses(Hjava_lang_String *);
 static HObject *compilerCommand(HObject *);
 
-#define INITIALIZE_METHOD(MB) \
-  if (OPT_SETQ(OPT_CMPLCLINIT) || strcmp((MB)->fb.name, "<clinit>")) {\
-    (MB)->invoker = sym_compileAndInvokeMethod;\
-    (MB)->CompiledCode = NULL;\
-  }
-
 
 /*
  * Initialization of the compiler.
@@ -158,8 +142,6 @@ void java_lang_Compiler_start(
   JITInterface *jitinterface
 #endif
 ) {
-  static int already_called = 0;
-
   ExecEnv *ee = EE();
   MONITOR_T key;
   int version;
@@ -170,18 +152,18 @@ void java_lang_Compiler_start(
   fflush(stdout);
 #endif
 
-  key = obj_monitor((JHandle *)classJavaLangClass);
-  monitorEnter2(ee, key);
-
-  if (already_called) {
-    monitorExit2(ee, key);
-    return;
+  // prevent unloading of java.lang.Compiler class
+  {
+    ClassClass *compiler_cb = FindClass(ee, "java/lang/Compiler", TRUE);
+    if (!compiler_cb) {
+      printf("FATAL: cannot find the java.lang.Compiler class.\n");
+      JVM_Exit(1);
+    }
+    CCSet(compiler_cb, Sticky);
   }
-  already_called = 1;
-  monitorExit2(ee, key);
 
 
-  /* version check */
+  // version check
 
 #if JDK_VER < 12
   version = *(int *)vector[0];
@@ -204,10 +186,10 @@ void java_lang_Compiler_start(
     printf("Version num. of class file format: %d.%d\n",
 	javaVersion, javaMinorVersion);
   }
-#endif	/* COMPILE_DEBUG */
+#endif	// COMPILE_DEBUG
 
 
-  /* get options */
+  // get options
   {
     char *opts = getenv("JAVA_COMPILER_OPT");
     char *opt;
@@ -236,157 +218,40 @@ void java_lang_Compiler_start(
 	}
 	opt = strtok(NULL, ", ");
       }
+
+      fflush(stdout);
     }
   }
 
 
-  /* show credit */
+  // show credit
   if (!OPT_SETQ(OPT_QUIET)) {
     fprintf(stderr, CREDIT);
     fflush(stderr);
   }
 
 
-  /* initialization */
-
-  /* sysAddDLSegment() some libraries to get addresses of functions */
-  /* __divdi3, __moddi3, printf, fflush in libc */
-  /* fmod in libm */
+#ifdef FORCE_DOUBLE_PRECISION
+  // set rounding precision as double
   {
-    char *libs_orig = getenv("JAVA_COMPILER_LIBS");
-    if (!libs_orig)  libs_orig = LIBS_ADDDLSEG;	/* defined in compiler.h */
-
-    if (libs_orig) {
-      char *libs;
-      char sep, nextsep, *elem, *next;
-#define NEXT {\
-  char *commap, *barp;\
-  elem = next;  sep = nextsep;\
-  commap = strchr(elem, ',');  barp = strchr(elem, '|');\
-  if (commap)\
-    if (barp)\
-      if (barp < commap) { next = barp;  nextsep = '|'; }\
-      else { next = commap;  nextsep = ','; }\
-    else { next = commap;  nextsep = ','; }\
-  else\
-    if (barp) { next = barp;  nextsep = '|'; }\
-    else  { next = strchr(elem, '\0');  nextsep = '\0'; }\
-  *next = '\0';  next++;\
-}
-
-#define SKIP {\
-  char *commap;\
-  if (nextsep == ',')  NEXT\
-  else {\
-    commap = strchr(next, ',');\
-    if (commap) {\
-      next = commap + 1;  nextsep = ',';\
-      NEXT\
-    }\
-    else  sep = '\0';\
-  }\
-}
-
-      libs = strdup(libs_orig);
-
-      next = libs;  nextsep = '|';	/* initialize */
-      NEXT
-      do {
-	int loaded = 0;
-	do {
-	  struct stat buf;
-	  if (!stat(elem, &buf)) {
-	    if (JVM_LoadLibrary(elem)) {
-	      loaded = 1;
-	      break;
-	    }
-	  }
-#ifdef COMPILE_DEBUG
-	  else {
-	    printf("can't stat() file: %s\n", elem);  fflush(stdout);
-	  }
-#endif
-	  NEXT
-	} while (sep == '|');
-
-	if (!loaded) {
-	  printf("FATAL: JIT compiler can't find required libraries: %s\n",
-			libs_orig);
-	  JVM_Exit(1);
-	}
-
-	SKIP
-      } while (sep != '\0');
-
-      sysFree(libs);
-    }	/* if (libs_orig) */
+    unsigned short cw;
+    asm("fnstcw %0" : "=m"(cw));
+    cw = (cw & ~0x0300) | 0x0200;
+    asm("fldcw %0" : : "m"(cw));
   }
+#endif	// FORCE_DOUBLE_PRECISION
 
 
-#if defined(RUNTIME_DEBUG) || defined(COMPILE_DEBUG)
-  {
-    char fpuenv[28];
-    fillFPUEnv(fpuenv);
-    showFPUEnv(stderr, "  ", fpuenv);
-  }
-#endif
+  // prohibit generating lossy opcodes
+  UseLosslessQuickOpcodes = TRUE;
 
 
-#if (defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT))&& defined(SEARCH_SIGCONTEXT)
-#ifdef linux
-  asm("movw %%gs,%0" : "=m" (reg_gs));
-  asm("movw %%fs,%0" : "=m" (reg_fs));
-#endif
-  asm("movw %%es,%0" : "=m" (reg_es));
-  asm("movw %%ds,%0" : "=m" (reg_ds));
-#ifdef __FreeBSD__
-  asm("movw %%cs,%0" : "=m" (reg_cs));
-  asm("movw %%ss,%0" : "=m" (reg_ss));
-#endif
-
-#ifdef RUNTIME_DEBUG
-#ifdef linux
-  printf("segment registers (gs,fs,es,ds): %04x, %04x, %04x, %04x\n",
-	reg_gs, reg_fs, reg_es, reg_ds);
-#endif	/* linux */
-#ifdef __FreeBSD__
-  printf("segment registers (es,ds,cs,ss): %04x, %04x, %04x, %04x\n",
-	reg_es, reg_ds, reg_cs, reg_ss);
-#endif	/* __FreeBSD__ */
-  fflush(stdout);
-#endif
-#endif	/* (EXC_BY_SIGNAL || GET_SIGCONTEXT) && SEARCH_SIGCONTEXT */
-
-#ifdef EXC_BY_SIGNAL
-  /* prepare struct sigaction for signalHandler() */
-#ifdef SA_SIGINFO
-  sigActForHandler.sa_handler = NULL;
-  sigActForHandler.sa_sigaction =
-	(void (*)(int,siginfo_t *,void *))signalHandler;
-#else
-  sigActForHandler.sa_handler = (void (*)(int))signalHandler;
-#endif
-  sigemptyset(&sigActForHandler.sa_mask);
-  sigaddset(&sigActForHandler.sa_mask, SIGPIPE);
-  sigActForHandler.sa_flags = 0;
-#ifdef SA_SIGINFO
-  sigActForHandler.sa_flags |= SA_SIGINFO;
-#endif
-#ifdef SA_RESTART
-  sigActForHandler.sa_flags |= SA_RESTART;
-#endif
-#endif
-
-
-#ifdef RESOLVE_SYMBOL_ON_CODE
-  /* resolve function symbols in pre-assembled code (assembledCode()) */
-  initFunctionSymbols();
-#endif
-
-
-  /* resolve some symbols */
+  // resolve some symbols
 #define RESOLVE_A_INVOKER(NAME)	\
-	sym_##NAME = (void *)symbolInSystemClassLoader(#NAME)
+	if (!(sym_##NAME = (void *)symbolInSystemClassLoader(#NAME))) {\
+		printf("FATAL: cannot resolve a symbol: %s\n", #NAME);\
+		JVM_Exit(1);\
+	}
 
   RESOLVE_A_INVOKER(compileAndInvokeMethod);
   RESOLVE_A_INVOKER(invokeJITCompiledMethod);
@@ -417,42 +282,86 @@ void java_lang_Compiler_start(
 #endif
 
 
-  /* initialize the classes which is already loaded */
+  // initialize all classes already loaded
   if (!OPT_SETQ(OPT_DONTCMPLVMCLZ)) {
     ClassClass **clazzptr;
     int i;
 
-#if JDK_VER < 12
-    BINCLASS_LOCK();
-#else
+#if JDK_VER >= 12
     BINCLASS_LOCK(sysThreadSelf());
+#else
+    BINCLASS_LOCK();
 #endif
 #ifdef COMPILE_DEBUG
     printf("%d classes is already loaded.\n", nbinclasses);
     fflush(stdout);
 #endif
+
+    // initialize
     for (i = nbinclasses, clazzptr = binclasses; --i >= 0; clazzptr++) {
 #ifdef COMPILE_DEBUG
-      printf("classes already in VM: %s\n", cbName(*clazzptr));
+      printf("init. a class in VM: %x,%s\n", *clazzptr, cbName(*clazzptr));
       fflush(stdout);
 #endif
-      initializeClass(*clazzptr);
+      initializeClassForJIT(*clazzptr, TRUE);
     }
-#if JDK_VER < 12
-    BINCLASS_UNLOCK();
-#else
+
+    if (OPT_SETQ(OPT_CMPLATLOAD)) {
+      // compile
+      ClassClass **classes;
+      classes = (ClassClass **)sysMalloc(sizeof(ClassClass *) * nbinclasses);
+      memcpy(classes, binclasses, sizeof(ClassClass *) * nbinclasses);
+
+      clazzptr = classes;
+      for (i = nbinclasses, clazzptr = classes; --i >= 0; clazzptr++) {
+#ifdef COMPILE_DEBUG
+	printf("compile a class in VM: %x,%s\n", *clazzptr, cbName(*clazzptr));
+	fflush(stdout);
+#endif
+	compileClass(*clazzptr);
+      }
+
+      sysFree(classes);
+    }
+
+#if JDK_VER >= 12
     BINCLASS_UNLOCK(sysThreadSelf());
+#else
+    BINCLASS_UNLOCK();
 #endif
   }
 
 
-  /* prohibit generating lossy opcodes */
-  UseLosslessQuickOpcodes = TRUE;
+  // initialize for signal handler
+#if (defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)) && defined(SEARCH_SIGCONTEXT)
+#  if JDK_VER < 12
+  *(char **)vector[3] = (char *)examineSigcontextNestCount;
+#  else
+  *((JITInterface6 *)jitinterface)->p_CompiledCodeSignalHandler =
+	examineSigcontextNestCount;
+#  endif
+
+  asm(".global exam_point\n\t"
+      "exam_point: int $0x10");
+
+#  if JDK_VER < 12
+  *(char **)vector[3] = NULL;
+#  else
+  *((JITInterface6 *)jitinterface)->p_CompiledCodeSignalHandler = NULL;
+#  endif
+
+  {
+    if (sc_nest < 0) {
+      printf("FATAL: cannot examine the offset of signal context.\n");
+      JVM_Exit(1);
+    }
+  }
+#endif	// SEARCH_SIGCONTEXT
 
 
-  /* set up link vector */
+  // set up link vector
 #if JDK_VER < 12
-  *(char **)vector[1] = (char *)initializeClass;
+  *(char **)vector[1] = (char *)initializeClassHook;
   *(char **)vector[2] = (char *)sym_invokeJITCompiledMethod;
   *(char **)vector[3] = (char *)signalHandler;
   *(char **)vector[4] = (char *)freeClass;
@@ -463,10 +372,12 @@ void java_lang_Compiler_start(
     *(char **)vector[7] = (char *)compilerEnable;
     *(char **)vector[8] = (char *)compilerDisable;
   }
-#endif	/* IGNORE_DISABLE */
+#endif	// IGNORE_DISABLE
   *(char **)vector[10] = (char *)pcInCompiledCode;
   *(char **)vector[11] = (char *)compiledCodePC;
+#ifdef DIRECT_INVOCATION
   *(char **)vector[70] = (char *)framePrev;
+#endif	// DIRECT_INVOCATION
 #ifdef COMPILE_DEBUG
   {
     int i;
@@ -477,12 +388,12 @@ void java_lang_Compiler_start(
     printf("*vector[70]: 0x%08x\n", i, (int)*(char **)vector[70]);
     fflush(stdout);
   }
-#endif	/* COMPILE_DEBUG */
-#else	/* JDK_VER */
+#endif	// COMPILE_DEBUG
+#else	// JDK_VER
   {
     JITInterface6 *ji6 = (JITInterface6 *)jitinterface;
 
-    *ji6->p_InitializeForCompiler = initializeClass;
+    *ji6->p_InitializeForCompiler = initializeClassHook;
     *ji6->p_invokeCompiledMethod = sym_invokeJITCompiledMethod;
     *ji6->p_CompiledCodeSignalHandler = signalHandler;
     *ji6->p_CompilerFreeClass = freeClass;
@@ -494,17 +405,20 @@ void java_lang_Compiler_start(
       *ji6->p_CompilerEnable = compilerEnable;
       *ji6->p_CompilerDisable = compilerDisable;
     }
-#endif	/* IGNORE_DISABLE */
+#endif	// IGNORE_DISABLE
 
     *ji6->p_PCinCompiledCode = pcInCompiledCode;
     *ji6->p_CompiledCodePC = compiledCodePC;
+#ifdef DIRECT_INVOCATION
     *ji6->p_CompiledFramePrev = framePrev;
+#endif	// DIRECT_INVOCATION
+    *ji6->p_CompiledFrameID = frameID;	// only for JDK >= 1.2
   }
-#endif	/* JDK_VER */
+#endif	// JDK_VER
 
 
 #ifdef CODE_DB
-  /* open DB and page file */
+  // open DB and page file
   if (OPT_SETQ(OPT_CODEDB)) {
     void *dl_dbm;
     if (!(dl_dbm = dlopen(LIBDBM, RTLD_LAZY))) {
@@ -559,21 +473,34 @@ void java_lang_Compiler_start(
     JVM_Exit(1);
   codedb_init_done:
   }
-#endif	/* CODE_DB */
+#endif	// CODE_DB
 
 
 #ifdef METAVM
   {
     ExecEnv *ee = EE();
-    /* force JVM to load Proxy class */
+    // force JVM to load Proxy class
     FindClass(ee, METAVM_PKG "Proxy", TRUE);
 
-    /* set remote flag on */
+    // set remote flag on
     REMOTE_FLAG_ON(ee);
   }
 #endif
 
 
+  // for strictfp
+  // check the FPU roundig precision
+  {
+    uint16_t cw;
+    asm("fstcw %0" : "=m" (cw));
+    is_fpupc_double = ((cw & 0x0300) == 0x0200);
+#if defined(RUNTIME_DEBUG) || defined(COMPILE_DEBUG)
+    printf("FPU rounding precision: %sdouble.\n",
+	(is_fpupc_double?"":"not "));
+#endif
+  }
+
+
 #ifdef COMPILE_DEBUG
   printf("java_lang_Compiler_start() done.\n");
   fflush(stdout);
@@ -581,22 +508,15 @@ void java_lang_Compiler_start(
 }
 
 
-/*
- * Local Functions
- */
-
 /*
  * Initialize the class at the time one is loaded.
  */
-void initializeClass(ClassClass *cb) {
-  struct methodblock *mb, *mb_end;
-#ifdef COMPILE_DEBUG
-  printf("initializeClass(%s) called.\n", cbName(cb));
-  fflush(stdout);
-#endif
+void initializeClassForJIT(ClassClass *cb, bool_t initInvoker) {
+  struct methodblock *mb;
+  int mb_count;
 
 #ifdef METAVM
-  /* force a loaded class to implement java.io.Serializable */
+  // force a loaded class to implement java.io.Serializable
   {
     ExecEnv *ee = EE();  JNIEnv *env = EE2JNIEnv(ee);
     static ClassClass *clz_Serializable = NULL;
@@ -611,373 +531,57 @@ void initializeClass(ClassClass *cb) {
   }
 #endif
 
-  if (OPT_SETQ(OPT_CMPLATLOAD)) {
-    compileClass(cb);
-    return;
-  }
-
   mb = cbMethods(cb);
-  mb_end = mb + cbMethodsCount(cb);
-  for (; mb < mb_end; mb++) {
+  mb_count = cbMethodsCount(cb);
+  for (; mb_count-- > 0; mb++) {
     int access = mb->fb.access;
 
     if (access & ACC_ABSTRACT)  continue;
 
-    /* initialize mb->CompiledCodeInfo */
-    if (!(mb->CompiledCodeInfo))  prepareCompiledCodeInfo(EE(), mb);
+    // initialize mb->CompiledCodeInfo
+    if (!(mb->CompiledCodeInfo))
+      if (!prepareCompiledCodeInfo(EE(), mb))  continue;
 
     if (access & ACC_NATIVE)  continue;
 
-    INITIALIZE_METHOD(mb);
+    if (!initInvoker)  continue;
+
+    // initialize invoker
+    if ((OPT_SETQ(OPT_CMPLCLINIT) || strcmp(mb->fb.name, "<clinit>")) &&
+	(mb->invoker != sym_compileAndInvokeMethod) &&	// not init.ed yet
+	(mb->invoker != sym_invokeJITCompiledMethod)) {	// not compiled yet
+      mb->invoker = sym_compileAndInvokeMethod;
+      mb->CompiledCode = NULL;
+    }
   }
+}
+
+
+//
+// Local Functions
+//
+
+static void initializeClassHook(ClassClass *cb) {
 #ifdef COMPILE_DEBUG
-  printf("initializeClass(%s) done.\n", cbName(cb));
+  printf("initializeClassHook(%s) called.\n", cbName(cb));
   fflush(stdout);
 #endif
-}
 
+  initializeClassForJIT(cb, TRUE);
 
-#ifdef __FreeBSD__
-static bool_t signalHandler(int sig, int code, struct sigcontext *uc) {
-#else
-static bool_t signalHandler(int sig, void *info, void *uc) {
-#endif
-#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
-#  ifdef SEARCH_SIGCONTEXT
-  static int sc_offset = -1;
-  static int sc_search_times = 2;
-#if 1
-  int32_t *ptr = ((int32_t *)&sig) - 1;
-#else
-  int32_t *ptr = ((int32_t *)&uc);
-#endif
-#  endif	/* SEARCH_SIGCONTEXT */
-  SIGCONTEXT *sc = NULL;
-  struct methodblock *mb;
-  CodeInfo *codeinfo;
-  uint32_t native_off;
-#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */
-#ifdef EXC_BY_SIGNAL
-  throwentry *tentry;
-  uint32_t jump_target = 0;
-#endif	/* EXC_BY_SIGNAL */
+  if (OPT_SETQ(OPT_CMPLATLOAD)) {	// have to done after initialization
+    compileClass(cb);
+  }
 
-#if defined(RUNTIME_DEBUG) || defined(COMPILE_DEBUG)
-  printf("Signal handler: %d (sys_thread_t: %x)\n", sig, (int)sysThreadSelf());
-  showStackFrames(EE());
-#endif
-
-
-#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
-  /* get methodblock */
-#if 1
-  mb = EE()->current_frame->current_method;
-	/* EBP is broken on FreeBSD ??? */
-#else
-  mb = (struct methodblock *) *(((int32_t *)sc->SC_EBP) + 3);
-	/* mb is second argument of assembledCode(): 0xc(%ebp) */
-#endif
-#ifdef RUNTIME_DEBUG
-  printf("method(0x%x)", (int)mb);  fflush(stdout);
-  fflush(stdout);
-  if (!mb)  goto signal_handler_error;
-  printf(": %s#%s %s\n",
-	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
+#ifdef COMPILE_DEBUG
+  printf("initializeClassHook(%s) done.\n", cbName(cb));
   fflush(stdout);
 #endif
-  codeinfo = (CodeInfo *)(mb->CompiledCodeInfo);
-
-
-#ifndef SEARCH_SIGCONTEXT
-  sc = (SIGCONTEXT *)uc;
-#else
-  /* get signal context */
-  if (--sc_search_times < 0) {
-    sc = (SIGCONTEXT *)(ptr + sc_offset);
-  }
-  else {	/* must search */
-#define SC_SEARCH_WIDTH	50
-    for (sc_offset = 0; sc_offset < SC_SEARCH_WIDTH - 3; sc_offset++) {
-#if defined(linux)
-      if ( (*((unsigned short *)(ptr)) == reg_gs) &&
-	(*((unsigned short *)(ptr + 1)) == reg_fs) &&
-	(*((unsigned short *)(ptr + 2)) == reg_es) &&
-	(*((unsigned short *)(ptr + 3)) == reg_ds) )
-	break;
-#elif defined(__FreeBSD__)
-      if ( (*((unsigned short *)(ptr)) == reg_es) &&
-	(*((unsigned short *)(ptr + 1)) == reg_ds) &&
-	(*((unsigned short *)(ptr + 2)) == reg_cs) &&
-	(*((unsigned short *)(ptr + 3)) == reg_ss) ) {
-	sc_offset -= 7;
-	ptr -= 7;
-	break;
-      }
-#else
-      if (0)  break;
-#endif
-      else
-	ptr += 1;
-    }
-#ifdef RUNTIME_DEBUG
-    printf("sc_offset: %d\n", sc_offset);
-#endif
-    if (sc_offset >= (SC_SEARCH_WIDTH - 3)) {
-      printf("FATAL: cannot find struct sigcontext on stack.");
-      JVM_Exit(1);
-    }
-    sc = (SIGCONTEXT *)ptr;
-  }
-#endif	/* SEARCH_SIGCONTEXT*/
-
-#if defined(RUNTIME_DEBUG) || (defined(GET_SIGCONTEXT) && !defined(EXC_BY_SIGNAL))
-  printf("sigcontext: 0x%08x\n", (int)sc);  fflush(stdout);
-  showSigcontext(sc);
-#endif
-#endif	/* EXC_BY_SIGNAL */
-
-
-#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
-  native_off =
-	((uint32_t)sc->SC_EIP) - ((uint32_t)mb->CompiledCode);
-#ifdef RUNTIME_DEBUG
-  printf("native off: 0x%x\n", native_off);
-  fflush(stdout);
-#endif
-#ifndef EXC_BY_SIGNAL
-  JVM_Exit(1);
-#else
-  tentry = throwtableGet(codeinfo, native_off);
-#ifdef RUNTIME_DEBUG
-  printf("throwentry: ");
-  fflush(stdout);
-  if (tentry)
-    printf("start: 0x%x, len: 0x%x, byteoff: 0x%x\n",
-		tentry->start, tentry->len, tentry->byteoff);
-  else
-    printf(" (null)\n");
-  fflush(stdout);
-#endif	/* RUNTIME_DEBUG */
-
-#define IP_SEARCH_WIDTH	50
-  if (!tentry) {	/* search EIP in native code */
-    uint32_t *sp = (uint32_t *)sc->SC_ESP;
-    uint32_t *bound = sp + IP_SEARCH_WIDTH;
-
-    while (++sp <= bound) {
-      native_off = *sp - (uint32_t)mb->CompiledCode;
-      if ((native_off >= 0) && (native_off <= codeinfo->code_size)) {
-	tentry = throwtableGet(codeinfo, native_off);
-	if (tentry) {
-#ifdef ARITHEXC_BY_SIGNAL
-	  if (sig == SIGFPE) {
-	    int opc = tentry->opcode;
-	    if ((opc != opc_idiv) && (opc != opc_irem) &&
-		(opc != opc_ldiv) && (opc != opc_lrem))
-	      continue;
-	  }
-#endif	/* ARITHEXC_BY_SIGNAL */
-	  sc->SC_ESP = (uint32_t)(sp + 1);	/* stack top in the frame */
-	  sc->SC_EIP = *sp;			/* %eip on the native code */
-	  goto tentry_is_found;
-	}
-      }
-    }
-    goto signal_handler_error;
-  tentry_is_found:
-  }
-
-#ifdef NULLEXC_BY_SIGNAL
-  if ((sig == SIGSEGV) || (sig == SIGBUS)) {
-	/* NullPointerException occurred */
-    SignalError(NULL, JAVAPKG "NullPointerException", 0);
-	/* set ee->exception.exc, exception object */
-  }
-#endif
-#if defined(NULLEXC_BY_SIGNAL) && defined(ARITHEXC_BY_SIGNAL)
-  else
-#endif
-#ifdef ARITHEXC_BY_SIGNAL
-  if (sig == SIGFPE) {
-    if ((tentry->opcode == opc_idiv) || (tentry->opcode == opc_irem)) {
-      if (sc->SC_ECX /* divisor */ == 0) {
-	/* ArithmeticException occurred */
-	SignalError(NULL, JAVAPKG "ArithmeticException", "/ by zero");
-      }
-      else if ((sc->SC_EAX /* dividend */ == 0x80000000) &&
-	       (sc->SC_ECX /* divisor  */ == -1)) {
-	/* idiv insn. of x86 causes it: 0x80000000 / -1 */
-	sc->SC_EAX = 0x80000000;
-	sc->SC_EDX = 0;
-	jump_target = sc->SC_EIP + 2;	/* `+ 2' skips idiv `0xf7 0xXX' */
-      }
-      else {
-	float dividend, divisor;
-	memcpy(&dividend, &(sc->SC_EAX), 4);
-	memcpy(&divisor, &(sc->SC_ECX), 4);
-	printf("FATAL: SIGFPE occurred in %s: %e %s %e.\n",
-		((tentry->opcode == opc_idiv)?"idiv":"irem"),
-		(double)dividend, ((tentry->opcode == opc_idiv)?"/":"%"),
-		(double)divisor);
-	fflush(stdout);
-
-	goto signal_handler_error;
-      }
-    }
-    else if ((tentry->opcode == opc_ldiv) || (tentry->opcode == opc_lrem)) {
-      uint32_t *sp = (uint32_t *)sc->SC_ESP;
-      sc->SC_EBP = *(sp + 4);	/* *(sp + 4) is %ebp `back up'ed in ldiv */
-
-      /* dividend: (*(sp + 1) << 32) | *sp
-	 divisor : (*(sp + 3) << 32) | *(sp + 2) */
-
-      if (!((*(sp + 2)) | (*(sp + 3)))) {
-	/* ArithmeticException occurred */
-	SignalError(NULL, JAVAPKG "ArithmeticException", "/ by zero");
-      }
-#if 0
-      /* functions called by ldiv (__divdi3(),__moddi3()) don't cause SIGFPE */
-      else if ((*sp == 0) && (*(sp + 1) == 0x80000000) &&
-	       ((*(sp +2)  & *(sp + 3)) == 0xffffffff)) {
-      }
-#endif
-    }
-    else {
-      printf("FATAL: SIGFPE occured in %s.",
-	((tentry->opcode == opc_ldiv)?"ldiv":"lrem"));
-      fflush(stdout);
-
-      goto signal_handler_error;
-    }
-  }	/* if (sig == SIGFPE) */
-#endif	/* ARITHEXC_BY_SIGNAL */
-#if defined(NULLEXC_BY_SIGNAL) || defined(ARITHEXC_BY_SIGNAL)
-  else
-#endif
-  {
-    return FALSE;
-  }
-
-
-  /* re-initialize signal handler
-     associated with raised signal */
-#ifdef __FreeBSD__
-  {
-    struct sigaction sigAct;
-    sigAct = sigActForHandler;
-    sigaction(sig, &sigAct, (struct sigaction *)NULL);
-  }
-#endif	/* __FreeBSD__ */
-
-#ifdef __FreeBSD__
-  /* unmask the signal */
-  {
-    sigset_t sigmask;
-    sigemptyset(&sigmask);
-    sigaddset(&sigmask, sig);
-    sigprocmask(SIG_UNBLOCK, &sigmask, NULL);
-  }
-#endif	/* __FreeBSD__ */
-
-
-  /* set (int32_t)bytepcoff (local variable in assembledCode) */
-  *(((int32_t *)sc->SC_EBP) - 1) = tentry->byteoff;
-
-
-  /* JDK 1.2: set program counter to jump to exception handler */
-  /* JDK 1.1: jump to exception handler */
-#  if JDK_VER >= 12
-  {
-    uint32_t exc_handler;
-    exc_handler = jump_target ? jump_target :
-		(uint32_t)(mb->CompiledCode + codeinfo->exc_handler_nativeoff);
-    sc->SC_EIP = exc_handler;
-  }
-
-  return TRUE;
-#  else
-  {
-    register uint32_t exc_handler asm("edi");
-    register SIGCONTEXT *sc_ptr asm("edx");
-
-    exc_handler = jump_target ? jump_target :
-		(uint32_t)(mb->CompiledCode + codeinfo->exc_handler_nativeoff);
-    sc_ptr = sc;
-    asm("movl %0,%%esi" : : "m" (sc_ptr->SC_ESI));
-    asm("movl %0,%%ebp" : : "m" (sc_ptr->SC_EBP));
-    asm("movl %0,%%esp" : : "m" (sc_ptr->SC_ESP));
-
-    asm("movl %0,%%eax" : : "m" (sc_ptr->SC_EAX));
-    asm("movl %0,%%edx" : : "m" (sc_ptr->SC_EDX));
-
-    asm("jmp  *%0" : : "r" (exc_handler));
-  }
-#  endif	/* JDK_VER */
-	/* jump to exception handler */
-
-#endif	/* ! EXC_BY_SIGNAL */
-#endif	/* EXC_BY_SIGNAL || GET_SIGCONTEXT */
-
-signal_handler_error:
-  fprintf(stderr, "Signal %d\n", sig);
-  fflush(stderr);
-#if defined(EXC_BY_SIGNAL) || defined(GET_SIGCONTEXT)
-  if (sc) {
-    struct methodblock *sig_mb = methodByPC((unsigned char *)sc->SC_EIP);
-    fprintf(stderr, "  in ");
-    if (sig_mb) {
-      uint32_t nativepc =
-		(uint32_t)sc->SC_EIP - (uint32_t)sig_mb->CompiledCode;
-      CodeInfo *info = (CodeInfo *)sig_mb->CompiledCodeInfo;
-      fprintf(stderr, "%s#%s %s\n",
-		cbName(fieldclass(&sig_mb->fb)),
-		sig_mb->fb.name, sig_mb->fb.signature);
-      fprintf(stderr, "  native PC: 0x%x(%d)\n", nativepc, nativepc);
-      if (info) {
-	throwentry *tentry = throwtableGet(info, nativepc);
-	if (tentry) {
-	  int bytepc = tentry->byteoff;
-	  fprintf(stderr, "  byte   PC: 0x%x(%d)\n", bytepc, bytepc);
-	}
-      }
-    }
-    {
-      char fpuenv[28];
-      fillFPUEnv(fpuenv);
-      showFPUEnv(stderr, "  ", fpuenv);
-    }
-    fprintf(stderr, "\n");
-    fflush(stderr);
-
-    showSigcontext(sc);
-
-    fprintf(stderr, "\n");
-    fflush(stderr);
-  }
-#endif
-  return FALSE;
-}
-
-
-static void showFPUEnv(FILE *fp, char *indent, char fpuenv[28]) {
-  if (!indent)  indent = "";
-  if (!fp)  fp = stdout;
-
-  fprintf(fp, "%sFPU control word: 0x%04x\n", indent,
-		*((int16_t *)(fpuenv + 0)));
-  fprintf(fp, "%sFPU status  word: 0x%04x\n", indent,
-		*((int16_t *)(fpuenv + 4)));
-  fprintf(fp, "%sFPU tag     word: 0x%04x\n", indent,
-		*((int16_t *)(fpuenv + 8)));
-  fprintf(fp, "%sFPU FIP         : 0x%08x\n", indent,
-		*((int32_t *)(fpuenv + 12)));
-
-  fflush(fp);
 }
 
 
 /*
- * Freeing class staffs related to the compiler.
+ * Freeing class stuffs related to the compiler.
  */
 static void freeClass(ClassClass *cb) {
   struct methodblock *mb = cbMethods(cb);
@@ -994,16 +598,21 @@ static void freeClass(ClassClass *cb) {
 
 
 static unsigned char *compiledCodePC(JavaFrame *frame,struct methodblock *mb) {
+#ifdef DIRECT_INVOCATION
+  // in case of DIRECT_INVOCATION, frame->lastpc is not up-to-date.
+  return mb->code;
+#else
   return (frame ? frame->lastpc : mb->code);
+#endif	// DIRECT_INVOCATION
 }
 
 
 static bool_t pcInCompiledCode(
 	unsigned char *pc /* mb->lastpc */, struct methodblock *mb) {
-#if 1	/* mb->lastpc is a pc on bytecode */
+#if 1	// mb->lastpc is a pc on bytecode
   unsigned int off = pc - mb->code;
   return (off < mb->code_length);
-#else	/* mb->lastpc is a pc on compiled code */
+#else	// mb->lastpc is a pc on compiled code
   CodeInfo *info = (CodeInfo *)mb->CompiledCodeInfo;
   if (info) {
     unsigned char *compiled_code = (unsigned char *)mb->CompiledCode;
@@ -1017,9 +626,69 @@ static bool_t pcInCompiledCode(
 }
 
 
+#ifdef DIRECT_INVOCATION
+#undef FRAMEPREV_DEBUG
 static JavaFrame *framePrev(JavaFrame *frame, JavaFrame *frame_buf) {
-  return frame->prev;
+  uint32_t *basep;
+  unsigned char *ret_addr;
+  extern void invokeJIT_compiled_done(), candi_compiled_done();
+
+#ifdef FRAMEPREV_DEBUG
+{
+  struct methodblock *mb;
+  printf("	| framePrev() called.\n");
+  mb = frame->current_method;
+  printf("	|   %s#%s %s\n",
+	cbName(fieldclass(&mb->fb)), mb->fb.name, mb->fb.signature);
+  fflush(stdout);
 }
+#endif
+  basep = (uint32_t *)(frame->vars);
+#ifdef FRAMEPREV_DEBUG
+printf("	| basep: %x\n", basep);
+fflush(stdout);
+#endif
+
+  if (frame != frame_buf) {	// first time
+    memset((void *)frame_buf, 0, sizeof(JavaFrame));
+    frame_buf->prev = frame->prev;
+  }
+
+  ret_addr = (unsigned char *)basep[1];
+  if ((ret_addr == (unsigned char *)invokeJIT_compiled_done) ||
+      (ret_addr == (unsigned char *)candi_compiled_done)) {
+#ifdef FRAMEPREV_DEBUG
+printf("	| limit.\n\n");
+fflush(stdout);
+#endif
+    return frame->prev;
+  }
+
+  basep = (uint32_t *)basep[0];
+  frame_buf->vars = (stack_item *)basep;
+  frame_buf->current_method = (struct methodblock *)basep[3];
+#ifdef FRAMEPREV_DEBUG
+{
+  struct methodblock *mb;
+  mb = frame_buf->current_method;
+  printf("	| method: %x\n", mb);
+  fflush(stdout);
+  printf("	|   buf->cur_method: %s#%s %s\n\n",
+	 cbName(fieldclass(&mb->fb)), mb->fb.name, mb->fb.signature);
+  fflush(stdout);
+}
+#endif
+  return frame_buf;
+}
+#endif	// DIRECT_INVOCATION
+
+
+#if JDK_VER >= 12
+static void *frameID(JavaFrame *frame) {
+  return (void *)frame->vars;
+}
+#endif
+
 
 
 #ifndef IGNORE_DISABLE
@@ -1042,7 +711,7 @@ static void compilerDisable() {
 #endif
   compiler_enabled = FALSE;
 }
-#endif	/* IGNORE_DISABLE */
+#endif	// IGNORE_DISABLE
 
 
 /*
@@ -1065,7 +734,7 @@ static bool_t compileClass(ClassClass *cb) {
 #endif
     return TRUE;
   }
-#endif	/* IGNORE_DISABLE */
+#endif	// IGNORE_DISABLE
 
   mb = cbMethods(cb);
   mb_end = mb + cbMethodsCount(cb);
@@ -1074,21 +743,11 @@ static bool_t compileClass(ClassClass *cb) {
 
     if (access & ACC_ABSTRACT)  continue;
 
-    /* initialize mb->CompiledCodeInfo */
+    // initialize mb->CompiledCodeInfo
     if (!(mb->CompiledCodeInfo))  prepareCompiledCodeInfo(EE(), mb);
 
     if (access & ACC_NATIVE)  continue;
 
-    if (!(mb->code)) {
-#ifdef COMPILE_DEBUG
-      printf("mb->code is NULL: %s#%s %s\n",
-		cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
-      fflush(stdout);
-#endif
-      INITIALIZE_METHOD(mb);
-      continue;
-    }
-
     if (compileMethod(mb))  return FALSE;
   }
 
@@ -1156,7 +815,7 @@ static HObject *compilerCommand(HObject *obj) {
   fflush(stdout);
 #endif
 
-  /* do nothing. */
+  // do nothing.
 
   return ret;
 }
